/********************************************************
 *
 * MPCDump.cpp
 *
 * Copyright (c) Jason Weiler 2004 - All Rights Reserved
 *
 ********************************************************/

#include <windows.h>
#include <stdio.h>

#define MIN(A,B)		(((A)<(B))?(A):(B))

typedef unsigned char uint8_t;
typedef char int8_t;
typedef unsigned short uint16_t;
typedef short int16_t;
typedef unsigned long uint32_t;
typedef long int32_t;


typedef enum {
	AFT_General=1,	// 1	// file extensions are part of the name
	AFT_Tile_Ent,// 2		// bmp
	AFT_Prop,	// 3		// unknown...perhaps bmp
	AFT_Tile,	// 4		// bmp
	AFT_Frame,	// 5		// bmp

	AFT_MAX
} EArchiveFileType_t;

typedef uint16_t Color16;

/* xxxxx--- -------- Red */
/* -----xxx xxx----- Green */
/* -------- ---xxxxx Blue */
#define GET_RED16(C)	(((C)>>11)&0x1f )
#define GET_GREEN16(C)	(((C)>>5)&0x003f )
#define GET_BLUE16(C)	(((C)>>0)&0x001f )

typedef struct
{
	char MagicNum[4];
	uint32_t CatalogOffset;
} ArchiveHeader_t;

typedef struct
{
	char Name[64];
	uint32_t Offset;
	uint32_t Size;
	uint32_t NotSure;	// same as size
	uint32_t Type;		// type
} ArchiveRecordItem_t;

static bool GetArchiveHeader(HANDLE hArch, ArchiveHeader_t  & ArchHdr);

static void CopyFileBytes(HANDLE hWriteFile, HANDLE hReadFile, uint32_t ByteCount);
static void ConvertRawToBmp(HANDLE hWriteFile, HANDLE hReadFile, uint32_t ByteCount);
static void ConvertTileToBmp(HANDLE hWriteFile, HANDLE hReadFile, uint32_t ByteCount);
static void DumpBmp(HANDLE hWriteFile, HANDLE hReadFile, uint32_t ByteCount, uint32_t Width, uint32_t Height);
static void Usage(const char * ProgramName);

int main (int argc, char *argv[])
{
	const char * DumpFolder = "mpcdump";

	if (2 > argc)
	{
		Usage(argv[0]);
		return 0;
	}
	else if (3 <= argc)
	{
		DumpFolder = argv[2];
	}

	if (FALSE == CreateDirectory(DumpFolder, NULL))
	{
		if (ERROR_ALREADY_EXISTS != GetLastError())
		{
			fprintf(stderr, "Couldn't create output folder \"%s\"\n", DumpFolder);
			return -1;
		}
	}


	ArchiveHeader_t ArchHdr = {0};
	ArchiveRecordItem_t ArchRec = {0};
	char outPathName[MAX_PATH] = "\0";

	HANDLE hArch=NULL, hReadFile=NULL, hWriteFile=NULL;

	hArch = CreateFile(argv[1], GENERIC_READ, FILE_SHARE_READ, NULL,
						OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	
	if (INVALID_HANDLE_VALUE == hArch)
	{
		DWORD foo = GetLastError();
		fprintf(stderr, "Couldn't open archive file: \"%s\"\n", argv[1]);
		return -2;
	}

	int rc = 0;

	DWORD numBytes = 0;
	uint32_t CatalogSize = 0;

	if (true == GetArchiveHeader(hArch, ArchHdr))
	{
		SetFilePointer(hArch, ArchHdr.CatalogOffset, NULL, FILE_BEGIN);

		ReadFile(hArch, &CatalogSize, sizeof(CatalogSize), &numBytes, NULL);

		// open a file pointer from which to read the archive=item data
		hReadFile = CreateFile(argv[1], GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if (INVALID_HANDLE_VALUE != hReadFile)
		{
			for (uint32_t i=0; i<CatalogSize; ++i)
			{

				ReadFile(hArch, &ArchRec, sizeof(ArchRec), &numBytes, NULL);
				printf("File: (0x%08X) Type:%X %s \n", ArchRec.Size, ArchRec.Type, ArchRec.Name);


				SetFilePointer(hReadFile, ArchRec.Offset, NULL, FILE_BEGIN);
				
				switch(ArchRec.Type)
				{
					case AFT_General:
					{
						// try to give this file an extension
						char * ext = strrchr(ArchRec.Name, '_');
						if (ext != NULL)
						{
							if ( (0==strcmpi(ext+1, "WAV"))
								|| (0==strcmpi(ext+1, "LUA"))
								|| (0==strcmpi(ext+1, "MSM"))
								|| (0==strcmpi(ext+1, "MP3")) )
							{
								*ext = '.';
							}
						}

						wsprintf(outPathName, "%s\\%s", DumpFolder, ArchRec.Name);

						hWriteFile = CreateFile(outPathName,
												GENERIC_READ | GENERIC_WRITE,
												FILE_SHARE_READ, NULL, CREATE_ALWAYS,
												FILE_ATTRIBUTE_NORMAL, NULL);

						CopyFileBytes(hWriteFile, hReadFile, ArchRec.Size);
						CloseHandle(hWriteFile);
					}
					break;

					case AFT_Tile_Ent:
					case AFT_Tile:
					{
						wsprintf(outPathName, "%s\\%s.bmp", DumpFolder, ArchRec.Name);
						hWriteFile = CreateFile(outPathName,
												GENERIC_READ | GENERIC_WRITE,
												FILE_SHARE_READ, NULL, CREATE_ALWAYS,
												FILE_ATTRIBUTE_NORMAL, NULL);
						ConvertTileToBmp(hWriteFile, hReadFile, ArchRec.Size);
						CloseHandle(hWriteFile);
						hWriteFile = NULL;
					}
					break;

					case AFT_Frame:
					{
						wsprintf(outPathName, "%s\\%s.bmp", DumpFolder, ArchRec.Name);
						hWriteFile = CreateFile(outPathName,
												GENERIC_READ | GENERIC_WRITE,
												FILE_SHARE_READ, NULL, CREATE_ALWAYS,
												FILE_ATTRIBUTE_NORMAL, NULL);
						ConvertRawToBmp(hWriteFile, hReadFile, ArchRec.Size);
						CloseHandle(hWriteFile);
						hWriteFile = NULL;
					}
					break;

					case AFT_Prop: // some sort of odds & ends file. I think it contains a font
					default:
					{
						// All others, ust raw-dump the file
						wsprintf(outPathName, "%s\\%s", DumpFolder, ArchRec.Name);
						hWriteFile = CreateFile(outPathName,
												GENERIC_READ | GENERIC_WRITE,
												FILE_SHARE_READ, NULL, CREATE_ALWAYS,
												FILE_ATTRIBUTE_NORMAL, NULL);
						CopyFileBytes(hWriteFile, hReadFile, ArchRec.Size);
						CloseHandle(hWriteFile);
					}
					break;

				}
			}

			printf("Done: %d files processed.\n", CatalogSize);
			rc = 0;

		}
		else
		{
			fprintf(stderr, "Error: Failed to open scanning file pointer.\n");
			rc = -4;
		}
	}
	else
	{
		fprintf(stderr, "Error: Failed to read archive header.\n");
		rc = -3;
	}
	

	CloseHandle(hReadFile);
	hReadFile = NULL;

	CloseHandle(hArch);
	hArch = NULL;

	return rc;
}

bool GetArchiveHeader(HANDLE hArch, ArchiveHeader_t & ArchHdr)
{
	DWORD NumBytes;

	SetFilePointer(hArch, 0, NULL, FILE_BEGIN);
	if (0 == ReadFile(hArch, &ArchHdr, sizeof(ArchHdr), &NumBytes, NULL))
	{
		return false;
	}

	return true;
}


void CopyFileBytes(HANDLE hWriteFile, HANDLE hReadFile, uint32_t ByteCount)
{
	uint8_t Buffer[1024];
	uint32_t readSize;
	DWORD numBytes;

	while (ByteCount != 0)
	{
		readSize = MIN(ByteCount, sizeof(Buffer));

		ReadFile(hReadFile, Buffer, readSize, &numBytes, NULL);
		WriteFile(hWriteFile, Buffer, readSize, &numBytes, NULL);

		ByteCount -= readSize;
	}
}

void ConvertTileToBmp(HANDLE hWriteFile, HANDLE hReadFile, uint32_t ByteCount)
{
	uint32_t Width;
	uint32_t Height;

	SetFilePointer(hReadFile, 4, NULL, FILE_CURRENT);
	SetFilePointer(hReadFile, 68, NULL, FILE_CURRENT);

	ByteCount -= 72;

	Width = 32;
	Height = (ByteCount/2)/Width;

	DumpBmp(hWriteFile, hReadFile, ByteCount, Width, Height);
}


void ConvertRawToBmp(HANDLE hWriteFile, HANDLE hReadFile, uint32_t ByteCount)
{
	uint32_t Width;
	uint32_t Height;
	DWORD numBytes;

	ReadFile(hReadFile, &Width, sizeof(Width), &numBytes, NULL);
	ReadFile(hReadFile, &Height, sizeof(Height), &numBytes, NULL);
	SetFilePointer(hReadFile, 68, NULL, FILE_CURRENT);

	ByteCount -= 76;

	DumpBmp(hWriteFile, hReadFile, ByteCount, Width, Height);

	return;
}


void DumpBmp(HANDLE hWriteFile, HANDLE hReadFile, uint32_t ByteCount, uint32_t Width, uint32_t Height)
{
	DWORD numBytes;
	int32_t iBytesPerLine = Width * 2;
	BITMAPFILEHEADER bfh = {0};
	BITMAPINFOHEADER bih = {sizeof(bih)};

	bfh.bfType = 'MB';
	bfh.bfSize = 0;
	bfh.bfOffBits = sizeof(bfh) + sizeof(bih);


	bih.biHeight = Height;
	bih.biWidth = Width;
	bih.biPlanes = 1;
    bih.biBitCount = 24;
    bih.biCompression = BI_RGB; 
    bih.biSizeImage = 0;
    bih.biXPelsPerMeter = 1000;
    bih.biYPelsPerMeter = 1000;
    bih.biClrUsed = 0;
    bih.biClrImportant = 0; 


	WriteFile(hWriteFile, &bfh, sizeof(bfh), &numBytes, NULL);

	WriteFile(hWriteFile, &bih, sizeof(bih), &numBytes, NULL);

	int32_t BackUpByteCount;
	uint32_t i, j;
	Color16 inColor;
	RGBTRIPLE outColor;
	uint8_t zeroByte = 0;

	/* back up two lines */
	BackUpByteCount = -(iBytesPerLine * 2);

	SetFilePointer(hReadFile, (Height-1) * Width * 2, NULL, FILE_CURRENT);

	for(i=0; i<Height; ++i)
	{
		for(j=0; j<Width; ++j)
		{
			ReadFile(hReadFile, &inColor, sizeof(inColor), &numBytes, NULL);
			outColor.rgbtRed = (GET_RED16(inColor)*256)/32;
			outColor.rgbtGreen = (GET_GREEN16(inColor)*256)/64;
			outColor.rgbtBlue = (GET_BLUE16(inColor)*256)/32;

			WriteFile(hWriteFile, &outColor, sizeof(outColor), &numBytes, NULL);
		}
	
		SetFilePointer(hReadFile, BackUpByteCount, NULL, FILE_CURRENT);

		/* pad output file to a DWORD */
		if ((Width*sizeof(outColor))&0x3)
		{
			for(j=4 - ((Width*sizeof(outColor))&0x3); j!=0; --j)
			{
				WriteFile(hWriteFile, &zeroByte, sizeof(zeroByte), &numBytes, NULL);
			}
		}
	}

	return;
}

void Usage(const char * ProgramName)
{
	fprintf(stderr, "Usage: %s InputFile.mpc [OutputDir]\n", ProgramName);
	fprintf(stderr, "     where: InputFile is a valid MPC archive file\n");
	fprintf(stderr, "            OutputDir is an optional output folder. \"mpcdump\" by default\n\n");
	
	return;
}